# Copyright (c) 2014-2015, Clemson University
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of Clemson University nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from django.db import models
from django.db.models.signals import post_save, post_delete, m2m_changed
from django.core.exceptions import ValidationError
from . import settings
import anyvcs
import os
import re
import shutil
import subprocess

VCS_CHOICES = (
  ('git', 'Git'),
  ('hg', 'Mercurial'),
  ('svn', 'Subversion'),
)
RIGHTS_CHOICES = (
  ('-', 'Deny'),
  ('r', 'Read-Only'),
  ('rw', 'Read-Write'),
)

name_rx = re.compile(r'^(?:[a-zA-Z0-9][a-zA-Z0-9_.+-]*/)*(?:[a-zA-Z0-9][a-zA-Z0-9_.+-]*)$')
hidden_path_rx = re.compile(r'(?:^|/)\.')

def makedirs(path):
  try:
    os.makedirs(path)
  except OSError as e:
    import errno
    if e.errno != errno.EEXIST:
      raise

def removedirs(path, stop=None):
  while path != stop:
    try:
      os.rmdir(path)
      path = os.path.dirname(path)
    except OSError, e:
      break

def post_save_proxy(sender, instance, **kwargs):
  instance.post_save(**kwargs)

def post_delete_proxy(sender, instance, **kwargs):
  instance.post_delete(**kwargs)


class Repo(models.Model):
  name = models.CharField(
    max_length = 100,
    unique = True,
    db_index = True,
  )
  path = models.CharField(
    max_length = 100,
    unique = True,
    blank = True,
    help_text = 'Either relative to VCSREPO_ROOT or absolute. Changing this will move the repository on disk.',
  )
  vcs = models.CharField(
    max_length = 3,
    default = 'git',
    choices = VCS_CHOICES,
    verbose_name = 'Version Control System',
  )
  public_read = models.BooleanField(
    default = False,
    verbose_name = 'Public Read Access',
  )
  created = models.DateTimeField(
    null = True,
    auto_now_add = True,
  )
  last_modified = models.DateTimeField(
    null = True,
    auto_now = True,
  )

  class Meta:
    db_table = 'anyvcs_repo'
    ordering = ['name']
    verbose_name = 'Repository'
    verbose_name_plural = 'Repositories'

  def __init__(self, *args, **kwargs):
    super(Repo, self).__init__(*args, **kwargs)
    self._old_path = self.path

  def __unicode__(self):
    return self.name

  @property
  def abspath(self):
    return os.path.join(settings.VCSREPO_ROOT, self.path)

  @property
  def repo(self):
    try:
      return self._repo
    except AttributeError:
      self._repo = anyvcs.open(self.abspath, self.vcs)
      return self._repo

  @property
  def public_rights(self):
    """Provided for backwards compatibility"""
    import warnings
    warn = 'public_rights is deprecated, use public_read instead'
    warnings.warn(warn, DeprecationWarning)
    if self.public_read:
      return 'r'
    else:
      return '-'

  @public_rights.setter
  def public_rights(self, value):
    import warnings
    warn = 'public_rights is deprecated, use public_read instead'
    warnings.warn(warn, DeprecationWarning)
    self.public_read = 'r' in value
    if 'w' in value:
      warn = 'write privilege for public_rights is ignored'
      warnings.warn(warn, RuntimeWarning)

  @property
  def ssh_uri(self):
    return self.get_uri('ssh')

  @property
  def anonymous_ssh_uri(self):
    return self.get_uri('anonymous-ssh')

  def get_uri(self, protocol):
    return self.format_uri(protocol, settings.VCSREPO_URI_CONTEXT) 

  def format_uri(self, protocol, context):
    context = dict(context)
    context['path'] = self.name
    fmt = settings.VCSREPO_URI_FORMAT[(self.vcs, protocol)]
    return fmt.format(**context)

  def post_save(self, created, **kwargs):
    if created:
      makedirs(self.abspath)
      self._repo = anyvcs.create(self.abspath, self.vcs)
    elif self._old_path != self.path:
      makedirs(os.path.dirname(self.abspath))
      old_abspath = os.path.join(settings.VCSREPO_ROOT, self._old_path)
      shutil.move(old_abspath, self.abspath)
      if not os.path.isabs(self._old_path):
        removedirs(os.path.dirname(old_abspath), settings.VCSREPO_ROOT)
    self._old_path = self.path
    if self.vcs == 'svn':
      self.update_svnserve()

  def post_delete(self, **kwargs):
    try:
      shutil.rmtree(self.abspath)
      removedirs(os.path.dirname(self.abspath), settings.VCSREPO_ROOT)
    except OSError as e:
      import errno
      if e.errno != errno.ENOENT:
        raise

  def relocate_path(self):
    if self.vcs == 'svn':
      self.path = os.path.join('svn', self.name)
    else:
      self.path = settings.VCSREPO_PATH_FUNCTION(self)

  def clean_fields(self, exclude=None):
    err = {}
    if not exclude or 'name' not in exclude:
      if not name_rx.match(self.name):
        err['name'] = ['Invalid name']
    if not exclude or 'path' not in exclude:
      if self.vcs == 'svn':
        self.path = os.path.join('svn', self.name)
      elif not self.path:
        self.path = settings.VCSREPO_PATH_FUNCTION(self)
      if hidden_path_rx.search(self.path):
        err['path'] = 'Invalid path'
      if settings.VCSREPO_CHECK_NESTED_PATHS:
        # verify we aren't nesting repo paths (e.g. a and a/b)
        # is this a parent of another repo? (is this the a for another a/b)
        if not (settings.VCSREPO_ALLOW_NESTED_PATHS or self.vcs == 'hg'):
          qs = type(self).objects.filter(path__startswith=self.path+'/')
          if qs.count() != 0:
            msg = 'This an ancestor of another repository which does not support nesting.'
            err.setdefault('path', []).append(msg)
        # is this a child of another repo? (is this the a/b for another a)
        updirs = []
        p = self.path
        while p and p != '/':
          p = os.path.dirname(p)
          updirs.append(p)
        qs = type(self).objects.filter(path__in=updirs)
        if settings.VCSREPO_ALLOW_NESTED_PATHS:
          qs = qs.exclude(vcs='hg')
        if qs.count() != 0:
          msg = 'This a subdirectory of another repository which does not support nesting.'
          err.setdefault('path', []).append(msg)
    if not exclude or 'vcs' not in exclude:
      if not filter(lambda x: x[0] == self.vcs, VCS_CHOICES):
        msg = 'Not a valid VCS type'
        err.setdefault('vcs', []).append(msg)
    if err:
      raise ValidationError(err)

  def update_svnserve(self):
    if self.vcs != 'svn':
      return
    import fcntl
    conf_path = os.path.join(self.abspath, 'conf', 'svnserve.conf')
    with open(conf_path, 'a') as conf:
      conf.seek(0)
      fcntl.lockf(conf, fcntl.LOCK_EX)
      conf.truncate()
      conf.write('[general]\n')
      if self.public_read:
        conf.write('anon-access = read\n')
      conf.write('authz-db = authz\n')
    authz_path = os.path.join(self.abspath, 'conf', 'authz')
    d = { '-': '' }
    with open(authz_path, 'a') as authz:
      authz.seek(0)
      fcntl.lockf(authz, fcntl.LOCK_EX)
      authz.truncate()
      if settings.VCSREPO_GROUP_ACL_FUNCTION:
        authz.write('[groups]\n')
        group_acl = settings.VCSREPO_GROUP_ACL_FUNCTION(self)
        for g in group_acl.keys():
          members = ','.join((u.username for u in g.user_set.all()))
          authz.write('@%s = %s\n' % (g.name, members))
      authz.write('\n[/]\n')
      if settings.VCSREPO_USER_ACL_FUNCTION:
        user_acl = settings.VCSREPO_USER_ACL_FUNCTION(self)
        for u, r in user_acl.iteritems(): 
          authz.write('%s = %s\n' % (u.username, d.get(r, r)))
      if settings.VCSREPO_USE_GROUP_RIGHTS:
        for g, r in group_acl.iteritems():
          authz.write('@%s = %s\n' % (g.name, d.get(r, r)))
      if self.public_read:
        authz.write('* = r\n')

# Repo signals
post_save.connect(post_save_proxy, dispatch_uid=__name__, sender=Repo)
post_delete.connect(post_delete_proxy, dispatch_uid=__name__, sender=Repo)


if settings.VCSREPO_USE_USER_RIGHTS:
  class UserRights(models.Model):
    repo = models.ForeignKey(
      Repo,
      db_index = True,
    )
    user = models.ForeignKey(
      settings.VCSREPO_USER_MODEL,
      db_index = True,
    )
    rights = models.CharField(
      max_length = 2,
      default = 'rw',
      choices = RIGHTS_CHOICES,
    )
    created = models.DateTimeField(
      null = True,
      auto_now_add = True,
    )
    last_modified = models.DateTimeField(
      null = True,
      auto_now = True,
    )

    class Meta:
      db_table = 'anyvcs_userrights'
      unique_together = ('repo', 'user')
      verbose_name = 'User Access Rights'
      verbose_name_plural = 'User Access Rights'

    def __unicode__(self):
      return u'%s/%s' % (self.repo, self.user)

    def post_save(self, created, **kwargs):
      self.repo.update_svnserve()
      self.repo.last_modified = self.last_modified
      self.repo.save()

    def post_delete(self, **kwargs):
      try:
        self.repo.update_svnserve()
        self.repo.last_modified = self.last_modified
        self.repo.save()
      except Repo.DoesNotExist:
        pass

  # UserRights signals
  post_save.connect(post_save_proxy, dispatch_uid=__name__, sender=UserRights)
  post_delete.connect(post_delete_proxy, dispatch_uid=__name__, sender=UserRights)


if settings.VCSREPO_USE_GROUP_RIGHTS:
  class GroupRights(models.Model):
    repo = models.ForeignKey(
      Repo,
      db_index = True,
    )
    group = models.ForeignKey(
      settings.VCSREPO_GROUP_MODEL,
      db_index = True,
    )
    rights = models.CharField(
      max_length = 2,
      default = 'rw',
      choices = RIGHTS_CHOICES,
    )
    created = models.DateTimeField(
      null = True,
      auto_now_add = True,
    )
    last_modified = models.DateTimeField(
      null = True,
      auto_now = True,
    )

    class Meta:
      db_table = 'anyvcs_grouprights'
      unique_together = ('repo', 'group')
      verbose_name = 'Group Access Rights'
      verbose_name_plural = 'Group Access Rights'

    def __unicode__(self):
      return u'%s/%s' % (self.repo, self.group)

    def post_save(self, created, **kwargs):
      self.repo.update_svnserve()
      self.repo.last_modified = self.last_modified
      self.repo.save()

    def post_delete(self, **kwargs):
      try:
        self.repo.update_svnserve()
        self.repo.last_modified = self.last_modified
        self.repo.save()
      except Repo.DoesNotExist:
        pass

  # GroupRights signals
  post_save.connect(post_save_proxy, dispatch_uid=__name__, sender=GroupRights)
  post_delete.connect(post_delete_proxy, dispatch_uid=__name__, sender=GroupRights)

  if (
    settings.VCSREPO_USER_MODEL == 'auth.User'
    and settings.VCSREPO_GROUP_MODEL == 'auth.Group'
  ):
    from django.contrib.auth.models import User

    def group_member_changed(instance, action, **kwargs):
      if action.startswith('post_'):
        qs = GroupRights.objects.filter(group=instance, repo__vcs='svn')
        for gr in qs.select_related('repo'):
          gr.repo.update_svnserve()

    m2m_changed.connect(group_member_changed, dispatch_uid=__name__, sender=User.groups.through)



# encoding: utf-8
import datetime
from south.db import db
from south.v2 import SchemaMigration
from django.db import models

class Migration(SchemaMigration):

    def forwards(self, orm):
        
        # Adding model 'Repo'
        db.create_table('anyvcs_repo', (
            ('id', self.gf('django.db.models.fields.AutoField')(primary_key=True)),
            ('name', self.gf('django.db.models.fields.CharField')(unique=True, max_length=100, db_index=True)),
            ('path', self.gf('django.db.models.fields.CharField')(unique=True, max_length=100, blank=True)),
            ('vcs', self.gf('django.db.models.fields.CharField')(default='git', max_length=3)),
            ('public_rights', self.gf('django.db.models.fields.CharField')(default='-', max_length=2)),
            ('created', self.gf('django.db.models.fields.DateTimeField')(auto_now_add=True, null=True, blank=True)),
            ('last_modified', self.gf('django.db.models.fields.DateTimeField')(auto_now=True, null=True, blank=True)),
        ))
        db.send_create_signal('django_anyvcs', ['Repo'])

        # Adding model 'UserRights'
        db.create_table('anyvcs_userrights', (
            ('id', self.gf('django.db.models.fields.AutoField')(primary_key=True)),
            ('repo', self.gf('django.db.models.fields.related.ForeignKey')(to=orm['django_anyvcs.Repo'])),
            ('user', self.gf('django.db.models.fields.related.ForeignKey')(to=orm['auth.User'])),
            ('rights', self.gf('django.db.models.fields.CharField')(default='rw', max_length=2)),
            ('created', self.gf('django.db.models.fields.DateTimeField')(auto_now_add=True, null=True, blank=True)),
            ('last_modified', self.gf('django.db.models.fields.DateTimeField')(auto_now=True, null=True, blank=True)),
        ))
        db.send_create_signal('django_anyvcs', ['UserRights'])

        # Adding unique constraint on 'UserRights', fields ['repo', 'user']
        db.create_unique('anyvcs_userrights', ['repo_id', 'user_id'])

        # Adding model 'GroupRights'
        db.create_table('anyvcs_grouprights', (
            ('id', self.gf('django.db.models.fields.AutoField')(primary_key=True)),
            ('repo', self.gf('django.db.models.fields.related.ForeignKey')(to=orm['django_anyvcs.Repo'])),
            ('group', self.gf('django.db.models.fields.related.ForeignKey')(to=orm['auth.Group'])),
            ('rights', self.gf('django.db.models.fields.CharField')(default='rw', max_length=2)),
            ('created', self.gf('django.db.models.fields.DateTimeField')(auto_now_add=True, null=True, blank=True)),
            ('last_modified', self.gf('django.db.models.fields.DateTimeField')(auto_now=True, null=True, blank=True)),
        ))
        db.send_create_signal('django_anyvcs', ['GroupRights'])

        # Adding unique constraint on 'GroupRights', fields ['repo', 'group']
        db.create_unique('anyvcs_grouprights', ['repo_id', 'group_id'])


    def backwards(self, orm):
        
        # Removing unique constraint on 'GroupRights', fields ['repo', 'group']
        db.delete_unique('anyvcs_grouprights', ['repo_id', 'group_id'])

        # Removing unique constraint on 'UserRights', fields ['repo', 'user']
        db.delete_unique('anyvcs_userrights', ['repo_id', 'user_id'])

        # Deleting model 'Repo'
        db.delete_table('anyvcs_repo')

        # Deleting model 'UserRights'
        db.delete_table('anyvcs_userrights')

        # Deleting model 'GroupRights'
        db.delete_table('anyvcs_grouprights')


    models = {
        'auth.group': {
            'Meta': {'object_name': 'Group'},
            'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'name': ('django.db.models.fields.CharField', [], {'unique': 'True', 'max_length': '80'}),
            'permissions': ('django.db.models.fields.related.ManyToManyField', [], {'to': "orm['auth.Permission']", 'symmetrical': 'False', 'blank': 'True'})
        },
        'auth.permission': {
            'Meta': {'ordering': "('content_type__app_label', 'content_type__model', 'codename')", 'unique_together': "(('content_type', 'codename'),)", 'object_name': 'Permission'},
            'codename': ('django.db.models.fields.CharField', [], {'max_length': '100'}),
            'content_type': ('django.db.models.fields.related.ForeignKey', [], {'to': "orm['contenttypes.ContentType']"}),
            'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'name': ('django.db.models.fields.CharField', [], {'max_length': '50'})
        },
        'auth.user': {
            'Meta': {'object_name': 'User'},
            'date_joined': ('django.db.models.fields.DateTimeField', [], {'default': 'datetime.datetime.now'}),
            'email': ('django.db.models.fields.EmailField', [], {'max_length': '75', 'blank': 'True'}),
            'first_name': ('django.db.models.fields.CharField', [], {'max_length': '30', 'blank': 'True'}),
            'groups': ('django.db.models.fields.related.ManyToManyField', [], {'to': "orm['auth.Group']", 'symmetrical': 'False', 'blank': 'True'}),
            'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'is_active': ('django.db.models.fields.BooleanField', [], {'default': 'True'}),
            'is_staff': ('django.db.models.fields.BooleanField', [], {'default': 'False'}),
            'is_superuser': ('django.db.models.fields.BooleanField', [], {'default': 'False'}),
            'last_login': ('django.db.models.fields.DateTimeField', [], {'default': 'datetime.datetime.now'}),
            'last_name': ('django.db.models.fields.CharField', [], {'max_length': '30', 'blank': 'True'}),
            'password': ('django.db.models.fields.CharField', [], {'max_length': '128'}),
            'user_permissions': ('django.db.models.fields.related.ManyToManyField', [], {'to': "orm['auth.Permission']", 'symmetrical': 'False', 'blank': 'True'}),
            'username': ('django.db.models.fields.CharField', [], {'unique': 'True', 'max_length': '30'})
        },
        'contenttypes.contenttype': {
            'Meta': {'ordering': "('name',)", 'unique_together': "(('app_label', 'model'),)", 'object_name': 'ContentType', 'db_table': "'django_content_type'"},
            'app_label': ('django.db.models.fields.CharField', [], {'max_length': '100'}),
            'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'model': ('django.db.models.fields.CharField', [], {'max_length': '100'}),
            'name': ('django.db.models.fields.CharField', [], {'max_length': '100'})
        },
        'django_anyvcs.grouprights': {
            'Meta': {'unique_together': "(('repo', 'group'),)", 'object_name': 'GroupRights', 'db_table': "'anyvcs_grouprights'"},
            'created': ('django.db.models.fields.DateTimeField', [], {'auto_now_add': 'True', 'null': 'True', 'blank': 'True'}),
            'group': ('django.db.models.fields.related.ForeignKey', [], {'to': "orm['auth.Group']"}),
            'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'last_modified': ('django.db.models.fields.DateTimeField', [], {'auto_now': 'True', 'null': 'True', 'blank': 'True'}),
            'repo': ('django.db.models.fields.related.ForeignKey', [], {'to': "orm['django_anyvcs.Repo']"}),
            'rights': ('django.db.models.fields.CharField', [], {'default': "'rw'", 'max_length': '2'})
        },
        'django_anyvcs.repo': {
            'Meta': {'object_name': 'Repo', 'db_table': "'anyvcs_repo'"},
            'created': ('django.db.models.fields.DateTimeField', [], {'auto_now_add': 'True', 'null': 'True', 'blank': 'True'}),
            'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'last_modified': ('django.db.models.fields.DateTimeField', [], {'auto_now': 'True', 'null': 'True', 'blank': 'True'}),
            'name': ('django.db.models.fields.CharField', [], {'unique': 'True', 'max_length': '100', 'db_index': 'True'}),
            'path': ('django.db.models.fields.CharField', [], {'unique': 'True', 'max_length': '100', 'blank': 'True'}),
            'public_rights': ('django.db.models.fields.CharField', [], {'default': "'-'", 'max_length': '2'}),
            'vcs': ('django.db.models.fields.CharField', [], {'default': "'git'", 'max_length': '3'})
        },
        'django_anyvcs.userrights': {
            'Meta': {'unique_together': "(('repo', 'user'),)", 'object_name': 'UserRights', 'db_table': "'anyvcs_userrights'"},
            'created': ('django.db.models.fields.DateTimeField', [], {'auto_now_add': 'True', 'null': 'True', 'blank': 'True'}),
            'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'last_modified': ('django.db.models.fields.DateTimeField', [], {'auto_now': 'True', 'null': 'True', 'blank': 'True'}),
            'repo': ('django.db.models.fields.related.ForeignKey', [], {'to': "orm['django_anyvcs.Repo']"}),
            'rights': ('django.db.models.fields.CharField', [], {'default': "'rw'", 'max_length': '2'}),
            'user': ('django.db.models.fields.related.ForeignKey', [], {'to': "orm['auth.User']"})
        }
    }

    complete_apps = ['django_anyvcs']



# Copyright (c) 2014-2015, Clemson University
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of Clemson University nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from django.http import HttpResponse, HttpResponseNotFound
from django.views.decorators.csrf import csrf_exempt
from django.views.decorators.http import require_http_methods
from django.core.exceptions import PermissionDenied
from . import models, settings
from .models import Repo
import json

try:
  from django.contrib.auth import get_user_model
except ImportError:
  from django.contrib.auth.models import User
  def get_user_model():
    return User

class DictEncoder(json.JSONEncoder):
  def default(self, o):
    if hasattr(o, '__dict__'):
      return o.__dict__
    else:
      return unicode(o)

dictencoder = DictEncoder()

def JsonResponse(data, *args, **kwargs):
  kwargs.setdefault('content_type', 'application/json')
  json = dictencoder.encode(data)
  return HttpResponse(json, *args, **kwargs)

def repo_access_data(repo, user):
  rights = settings.VCSREPO_RIGHTS_FUNCTION(repo, user)
  return { 'rights': rights, 'vcs': repo.vcs, 'path': repo.abspath }

def access(request, repo):
  username = request.GET.get('u')
  user = None
  if username:
    UserModel = get_user_model()
    try:
      user = UserModel.objects.get(username=username)
    except UserModel.DoesNotExist:
      message = 'User does not exist: %s\n' % username
      return HttpResponseNotFound(message, content_type='text/plain')

  vcs = request.GET.get('vcs')
  qs = Repo.objects.all()
  if vcs:
    qs = qs.filter(vcs=vcs)

  try:
    repo = qs.get(name=repo)
  except Repo.DoesNotExist:
    message = 'Repository does not exist: %s\n' % repo
    return HttpResponseNotFound(message, content_type='text/plain')
  data = repo_access_data(repo, user)
  return JsonResponse(data)

@csrf_exempt
@require_http_methods(["GET", "POST"])
def api_call(request, repo, attr):
  from collections import Callable
  try:
    repo = Repo.objects.get(name=repo)
  except Repo.DoesNotExist:
    message = 'Repository does not exist: %s\n' % repo
    return HttpResponseNotFound(message, content_type='text/plain')
  if not hasattr(repo.repo, attr):
    message = 'Attribute does not exist'
    return HttpResponseNotFound(message, content_type='text/plain')
  attr = getattr(repo.repo, attr)
  if request.method == 'POST':
    if not isinstance(attr, Callable):
      message = 'Attribute is not callable'
      return HttpResponseNotFound(message, content_type='text/plain')
    kwargs = json.load(request)
    try:
      data = attr(**kwargs)
    except Exception as e:
      import traceback
      data = {
        'module': type(e).__module__,
        'class': type(e).__name__,
        'args': e.args,
        'str': str(e),
        'traceback': traceback.format_exc(),
      }
      return JsonResponse(data, status=400)
    return JsonResponse(data)
  else:
    return JsonResponse(attr)



# Copyright (c) 2014-2015, Clemson University
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of Clemson University nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import httplib
import json
from urlparse import urlparse
from anyvcs.common import VCSRepo, PathDoesNotExist, BadFileType, attrdict, \
                          CommitLogEntry

class BadResponse(Exception):
  def __init__(self, status, reason, content_type, body=None):
    self.status = status
    self.reason = reason
    self.content_type = content_type
    self.body = body

  def __str__(self):
    return 'Server returned %s %s; Content-Type=%s' % (
      self.status, self.reason, self.content_type)

class VCSRepo(object):
  def __init__(self, api_url):
    url = urlparse(api_url)
    assert url.scheme == 'http'
    self._url = url
    self._conn = httplib.HTTPConnection(url.netloc)
    self._path = url.path.rstrip('/') + '/'

  def _getresponse(self):
    response = self._conn.getresponse()
    ct = response.getheader('Content-Type')
    body = response.read()
    if response.status == 200:
      if ct == 'application/json':
        data = json.loads(body)
        return data
    if response.status == 400:
      if ct == 'application/json':
        data = json.loads(body)
        if data.get('module') == 'anyvcs.common':
          klass = data.get('class')
          args = data.get('args', [])
          if klass == 'BadFileType':
            raise BadFileType(*args)
          if klass == 'PathDoesNotExist':
            raise PathDoesNotExist(*args)
    raise BadResponse(response.status, response.reason, ct, body)

  def _get(self, attr):
    method = 'GET'
    url = self._path + attr
    self._conn.request(method, url)
    return self._getresponse()

  def _post(self, attr, **kwargs):
    method = 'POST'
    url = self._path + attr
    body = json.dumps(kwargs)
    headers = {'Content-Type': 'application/json'}
    self._conn.request(method, url, body, headers)
    return self._getresponse()

  @property
  def path(self):
    return self._get('path')

  def ls(self, rev, path, recursive=False, recursive_dirs=False,
         directory=False, report=[]):
    if isinstance(report, (tuple, list)):
      report = ','.join(report)
    result = self._post('ls', rev=rev, path=path, recursive=recursive,
                        recursive_dirs=recursive_dirs, directory=directory,
                        report=','.join(report))
    return [attrdict(x) for x in result]

  def cat(self, rev, path):
    return self._post('cat', rev=rev, path=path)

  def readlink(self, rev, path):
    return self._post('readlink', rev=rev, path=path)

  def branches(self):
    return self._post('branches')

  def tags(self):
    return self._post('tags')

  def heads(self):
    return self._post('heads')

  def empty(self):
    return self._post('empty')

  def log(self, revrange=None, limit=None, firstparent=False, merges=None,
          path=None, follow=False):
    if isinstance(revrange, tuple):
      revrange = ','.join(revrange)
    result = self._post('log', revrange=revrange, limit=limit,
                        firstparent=firstparent, merges=merges, path=path,
                        follow=follow)
    if isinstance(result, list):
      return [CommitLogEntry(**x) for x in result]
    else:
      return CommitLogEntry(**result)

  def diff(self, rev_a, rev_b, path=None):
    return self._post('diff', rev_a=rev_a, rev_b=rev_b, path=path)

  def ancestor(self, rev1, rev2):
    return self._post('ancestor', rev1=rev1, rev2=rev2)

class GitRepo(VCSRepo):
  """Mirrors the functionality of anyvcs.git.GitRepo"""

  pass

class HgRepo(VCSRepo):
  """Mirrors the functionality of anyvcs.hg.HgRepo"""

  def bookmarks(self):
    return self._post('bookmarks')

class SvnRepo(VCSRepo):
  """Mirrors the functionality of anyvcs.svn.SvnRepo"""

  def proplist(self, rev, path=None):
    return self._post('proplist', rev=rev, path=path)

  def propget(self, prop, rev, path=None):
    return self._post('propget', prop=prop, rev=rev, path=path)

  def youngest(self):
    return self._post('youngest')



# -*- coding: utf-8 -*-
# Copyright (c) 2014-2015, Clemson University
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of Clemson University nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from django.test import TestCase
from django.test.client import Client
from django.http import Http404
from django.contrib.auth.models import User, Group
from django.core.exceptions import ValidationError
from django.core.urlresolvers import reverse
from django.utils.encoding import DjangoUnicodeDecodeError
from unittest import skipIf, skipUnless
from .models import Repo
from . import settings
from django_anyvcs import dispatch, shortcuts
import anyvcs.git, anyvcs.hg, anyvcs.svn
import gzip
import json
import os
import shutil
import subprocess
import tempfile

DEVNULL = open(os.devnull, 'wb')
GIT = 'git'

URI_FORMAT = {
  ('git', 'ssh'): "{user}@{hostname}:{path}",
  ('svn', 'ssh'): "svn+ssh://{user}@{hostname}/{path}",
  ('hg', 'ssh'): "ssh://{user}@{hostname}/{path}",
  ('git', 'anonymous-ssh'): "{anonymous}@{hostname}:{path}",
  ('svn', 'anonymous-ssh'): "svn+ssh://{anonymous}@{hostname}/{path}",
  ('hg', 'anonymous-ssh'): "ssh://{anonymous}@{hostname}/{path}",
}

URI_CONTEXT = {
  'anonymous': 'anonymous',
  'user': 'user',
  'hostname': 'hostname',
}

class BaseTestCase(TestCase):
  def setUp(self):
    self.original_root = settings.VCSREPO_ROOT
    settings.VCSREPO_ROOT = tempfile.mkdtemp(prefix='anyvcs-test.')
    self.original_uri_format = settings.VCSREPO_URI_FORMAT
    settings.VCSREPO_URI_FORMAT = URI_FORMAT
    self.original_uri_context = settings.VCSREPO_URI_CONTEXT
    settings.VCSREPO_URI_CONTEXT = URI_CONTEXT

  def tearDown(self):
    Repo.objects.all().delete()
    shutil.rmtree(settings.VCSREPO_ROOT)
    settings.VCSREPO_ROOT = self.original_root
    settings.VCSREPO_URI_FORMAT = self.original_uri_format
    settings.VCSREPO_URI_CONTEXT = self.original_uri_context

  def assertPathExists(self, path):
    if isinstance(path, (tuple, list)):
      path = os.path.join(*path)
    if not os.path.exists(path):
      raise AssertionError("Path does not exist: ", path)

  def assertPathNotExists(self, path):
    if isinstance(path, (tuple, list)):
      path = os.path.join(*path)
    try:
      self.assertPathExists(path)
    except AssertionError:
      return
    raise AssertionError("Path exists: ", path)

class TestingFrameworkTestCase(BaseTestCase):

  def test_assert_path_not_exists(self):
    path = tempfile.mkdtemp(prefix='django-anyvcs-path-exists')
    self.assertPathNotExists('/path/does/not/exist')
    self.assertPathNotExists(['/path', 'does', 'not', 'exist'])
    self.assertRaises(AssertionError, self.assertPathNotExists, path)
    self.assertRaises(AssertionError, self.assertPathNotExists, ['/'] + path.split(os.path.pathsep))
    os.rmdir(path)

  def test_assert_path_exists(self):
    path = tempfile.mkdtemp(prefix='django-anyvcs-path-exists')
    self.assertRaises(AssertionError, self.assertPathExists, '/path/does/not/exist')
    self.assertRaises(AssertionError, self.assertPathExists, ['/path', 'does', 'not', 'exist'])
    self.assertPathExists(path)
    self.assertPathExists(['/'] + path.split(os.path.pathsep))
    os.rmdir(path)

class FixtureTestCase(BaseTestCase):
  fixtures = ['django_anyvcs_basic.json']

  def setUp(self):
    # Can't call super() here because fixtures are loaded before setUp()
    self.repo = Repo.objects.get(name='repo')

  def tearDown(self):
    Repo.objects.all().delete()

  def test_path_exists(self):
    self.assertPathExists(self.repo.abspath)

class CreateRepoTestCase(BaseTestCase):
  def test_invalid_names(self):
    for name in ('$', '/', 'a//b'):
      repo = Repo(name=name, path='repo', vcs='git')
      self.assertRaises(ValidationError, repo.full_clean)

  def test_invalid_paths(self):
    for path in ('.', '..', 'a/..', '../a', '.a', '.a/a', 'a/.a'):
      repo = Repo(name='repo', path=path, vcs='git')
      self.assertRaises(ValidationError, repo.full_clean)

  def test_invalid_vcs(self):
    repo = Repo(name='repo', path='repo', vcs='invalid')
    self.assertRaises(ValidationError, repo.full_clean)

  def test_git(self):
    repo = Repo(name='a', path='b', vcs='git')
    repo.full_clean()
    repo.save()
    self.assertEqual(repo.path, 'b')
    self.assertEqual(repo.abspath, os.path.join(settings.VCSREPO_ROOT, 'b'))
    self.assertIsInstance(repo.repo, anyvcs.git.GitRepo)

  def test_git_without_path(self):
    repo = Repo(name='a', vcs='git')
    repo.full_clean()
    repo.save()
    self.assertIsInstance(repo.repo, anyvcs.git.GitRepo)

  def test_git_absolute_path(self):
    d = tempfile.mkdtemp(prefix='anyvcs-test.')
    path = os.path.join(d, 'repo')
    repo = Repo(name='a', path=path, vcs='git')
    repo.full_clean()
    repo.save()
    self.assertEqual(repo.path, path)
    self.assertEqual(repo.abspath, path)
    self.assertIsInstance(repo.repo, anyvcs.git.GitRepo)
    shutil.rmtree(d)

  def test_hg(self):
    repo = Repo(name='a', path='b', vcs='hg')
    repo.full_clean()
    repo.save()
    self.assertEqual(repo.path, 'b')
    self.assertEqual(repo.abspath, os.path.join(settings.VCSREPO_ROOT, 'b'))
    self.assertIsInstance(repo.repo, anyvcs.hg.HgRepo)

  def test_hg_without_path(self):
    repo = Repo(name='a', vcs='hg')
    repo.full_clean()
    repo.save()
    self.assertIsInstance(repo.repo, anyvcs.hg.HgRepo)

  def test_hg_absolute_path(self):
    d = tempfile.mkdtemp(prefix='anyvcs-test.')
    path = os.path.join(d, 'repo')
    repo = Repo(name='a', path=path, vcs='hg')
    repo.full_clean()
    repo.save()
    self.assertEqual(repo.path, path)
    self.assertEqual(repo.abspath, path)
    self.assertIsInstance(repo.repo, anyvcs.hg.HgRepo)
    shutil.rmtree(d)

  def test_svn(self):
    repo = Repo(name='a', path='b', vcs='svn')
    repo.full_clean()
    repo.save()
    self.assertEqual(repo.path, os.path.join('svn', 'a'))
    self.assertEqual(repo.abspath, os.path.join(settings.VCSREPO_ROOT, 'svn', 'a'))
    self.assertIsInstance(repo.repo, anyvcs.svn.SvnRepo)

  def test_svn_without_path(self):
    repo = Repo(name='a', vcs='svn')
    repo.full_clean()
    repo.save()
    self.assertEqual(repo.path, os.path.join('svn', 'a'))
    self.assertEqual(repo.abspath, os.path.join(settings.VCSREPO_ROOT, 'svn', 'a'))
    self.assertIsInstance(repo.repo, anyvcs.svn.SvnRepo)

  def test_svn_absolute_path(self):
    d = tempfile.mkdtemp(prefix='anyvcs-test.')
    path = os.path.join(d, 'repo')
    repo = Repo(name='a', path=path, vcs='svn')
    repo.full_clean()
    repo.save()
    self.assertEqual(repo.path, os.path.join('svn', 'a'))
    self.assertEqual(repo.abspath, os.path.join(settings.VCSREPO_ROOT, 'svn', 'a'))
    self.assertIsInstance(repo.repo, anyvcs.svn.SvnRepo)
    shutil.rmtree(d)

class ChangeRepoTestCase(BaseTestCase):
  def test_rename_git(self):
    repo = Repo(name='a', vcs='git')
    repo.full_clean()
    repo.save()
    old_path = repo.path
    repo.name = 'b'
    repo.full_clean()
    repo.save()
    self.assertEqual(repo.name, 'b')
    self.assertEqual(repo.path, old_path)
    self.assertPathExists(repo.abspath)

  def test_move_git(self):
    repo = Repo(name='a', vcs='git')
    repo.full_clean()
    repo.save()
    old_abspath = repo.abspath
    repo.path = 'b'
    repo.full_clean()
    repo.save()
    self.assertEqual(repo.name, 'a')
    self.assertEqual(repo.path, 'b')
    self.assertPathExists(repo.abspath)
    self.assertPathNotExists(old_abspath)

  def test_rename_svn(self):
    repo = Repo(name='a', vcs='svn')
    repo.full_clean()
    repo.save()
    repo.name = 'b'
    repo.full_clean()
    repo.save()
    self.assertEqual(repo.name, 'b')
    self.assertEqual(repo.path, os.path.join('svn', 'b'))
    self.assertPathExists(repo.abspath)

  def test_move_svn(self):
    repo = Repo(name='a', vcs='svn')
    repo.full_clean()
    repo.save()
    repo.path = 'b'
    repo.full_clean()
    repo.save()
    self.assertEqual(repo.name, 'a')
    self.assertEqual(repo.path, os.path.join('svn', 'a'))
    self.assertPathExists(repo.abspath)

class CannotDeleteSymlinkTestCase(BaseTestCase):
  def test(self):
    os.mkdir(os.path.join(settings.VCSREPO_ROOT, 'svn.target'))
    os.symlink('svn.target', os.path.join(settings.VCSREPO_ROOT, 'svn'))
    repo = Repo(name='a', vcs='svn')
    repo.full_clean()
    repo.save()
    repo.delete()

class LookupTestCase(BaseTestCase):
  @classmethod
  def setUpClass(cls):
    super(LookupTestCase, cls).setUpClass()
    cls.user1 = User.objects.create(username='user1')
    cls.user2 = User.objects.create(username='user2')
    cls.group1 = Group.objects.create(name='group1')
    cls.group2 = Group.objects.create(name='group2')
    cls.group1.user_set.add(cls.user1)
    cls.group2.user_set.add(cls.user2)

  @classmethod
  def tearDownClass(cls):
    User.objects.all().delete()
    Group.objects.all().delete()
    super(LookupTestCase, cls).tearDownClass()

  def test_public_read_anonymous(self):
    vcs = 'git'
    for public_read, public_rights in ((False, '-'), (True, 'r')):
      repo = Repo.objects.create(
        name = 'repo',
        path = 'repo',
        vcs = vcs,
        public_read = public_read,
      )
      client = Client()
      url = reverse('django_anyvcs.views.access', args=(repo.name,))
      response = client.get(url)
      self.assertEqual(response.status_code, 200)
      self.assertIn('Content-Type', response)
      self.assertEqual(response['Content-Type'], 'application/json')
      document = json.loads(response.content)
      self.assertIn('path', document)
      self.assertEqual(document['path'], os.path.join(settings.VCSREPO_ROOT, 'repo'))
      self.assertIn('vcs', document)
      self.assertEqual(document['vcs'], vcs)
      self.assertIn('rights', document)
      self.assertEqual(document['rights'], public_rights)
      repo.delete()

  def test_public_read_user(self):
    vcs = 'hg'
    for public_read, public_rights in ((False, '-'), (True, 'r')):
      repo = Repo.objects.create(
        name = 'repo',
        path = 'repo',
        vcs = vcs,
        public_read = public_read,
      )
      client = Client()
      url = reverse('django_anyvcs.views.access', args=(repo.name,))
      response = client.get(url, {'u': self.user1.username})
      self.assertEqual(response.status_code, 200)
      self.assertIn('Content-Type', response)
      self.assertEqual(response['Content-Type'], 'application/json')
      document = json.loads(response.content)
      self.assertIn('path', document)
      self.assertEqual(document['path'], os.path.join(settings.VCSREPO_ROOT, 'repo'))
      self.assertIn('vcs', document)
      self.assertEqual(document['vcs'], vcs)
      self.assertIn('rights', document)
      self.assertEqual(document['rights'], public_rights)
      repo.delete()

  @skipUnless(
    settings.VCSREPO_USE_USER_RIGHTS,
    "not using UserRights"
  )
  def test_user_overrides_public_read(self):
    from .models import UserRights
    vcs = 'git'
    for public_read, public_rights in ((False, '-'), (True, 'r')):
      for user_rights in ('-', 'r', 'rw'):
        repo = Repo.objects.create(
          name = 'repo',
          path = 'repo',
          vcs = vcs,
          public_read = public_read,
        )
        UserRights.objects.create(
          repo = repo,
          user = self.user1,
          rights = user_rights,
        )
        client = Client()
        url = reverse('django_anyvcs.views.access', args=(repo.name,))
        response = client.get(url, {'u': self.user1.username})
        self.assertEqual(response.status_code, 200)
        self.assertIn('Content-Type', response)
        self.assertEqual(response['Content-Type'], 'application/json')
        document = json.loads(response.content)
        self.assertIn('path', document)
        self.assertEqual(document['path'], os.path.join(settings.VCSREPO_ROOT, 'repo'))
        self.assertIn('vcs', document)
        self.assertEqual(document['vcs'], vcs)
        self.assertIn('rights', document)
        self.assertEqual(document['rights'], user_rights)
        repo.delete()

  @skipUnless(
    settings.VCSREPO_USE_GROUP_RIGHTS,
    "not using GroupRights"
  )
  def test_group_overrides_public_read(self):
    from .models import GroupRights
    vcs = 'hg'
    for public_read, public_rights in ((False, '-'), (True, 'r')):
      for group_rights in ('-', 'r', 'rw'):
        repo = Repo.objects.create(
          name = 'repo',
          path = 'repo',
          vcs = vcs,
          public_read = public_read,
        )
        GroupRights.objects.create(
          repo = repo,
          group = self.group1,
          rights = group_rights,
        )
        client = Client()
        url = reverse('django_anyvcs.views.access', args=(repo.name,))
        response = client.get(url, {'u': self.user1.username})
        self.assertEqual(response.status_code, 200)
        self.assertIn('Content-Type', response)
        self.assertEqual(response['Content-Type'], 'application/json')
        document = json.loads(response.content)
        self.assertIn('path', document)
        self.assertEqual(document['path'], os.path.join(settings.VCSREPO_ROOT, 'repo'))
        self.assertIn('vcs', document)
        self.assertEqual(document['vcs'], vcs)
        self.assertIn('rights', document)
        self.assertEqual(document['rights'], group_rights)
        repo.delete()

  @skipUnless(
    settings.VCSREPO_USE_USER_RIGHTS and settings.VCSREPO_USE_GROUP_RIGHTS,
    "not using UserRights and GroupRights"
  )
  def test_user_overrides_group_rights(self):
    from .models import UserRights, GroupRights
    vcs = 'git'
    for group_rights in ('-', 'r', 'rw'):
      for user_rights in ('-', 'r', 'rw'):
        repo = Repo.objects.create(
          name = 'repo',
          path = 'repo',
          vcs = vcs,
        )
        UserRights.objects.create(
          repo = repo,
          user = self.user1,
          rights = user_rights,
        )
        GroupRights.objects.create(
          repo = repo,
          group = self.group1,
          rights = group_rights,
        )
        client = Client()
        url = reverse('django_anyvcs.views.access', args=(repo.name,))
        response = client.get(url, {'u': self.user1.username})
        self.assertEqual(response.status_code, 200)
        self.assertIn('Content-Type', response)
        self.assertEqual(response['Content-Type'], 'application/json')
        document = json.loads(response.content)
        self.assertIn('path', document)
        self.assertEqual(document['path'], os.path.join(settings.VCSREPO_ROOT, 'repo'))
        self.assertIn('vcs', document)
        self.assertEqual(document['vcs'], vcs)
        self.assertIn('rights', document)
        self.assertEqual(document['rights'], user_rights)
        repo.delete()

  def test_rights_function(self):
    repo = Repo.objects.create(
      name = 'repo',
      path = 'repo',
      vcs = 'git',
    )
    original_rights_function = settings.VCSREPO_RIGHTS_FUNCTION
    try:
      for rights in ('-', 'r', 'rw'):
        def f(r, u):
          self.assertEqual(repo, r)
          self.assertEqual(self.user1, u)
          return rights
        settings.VCSREPO_RIGHTS_FUNCTION = f
        client = Client()
        url = reverse('django_anyvcs.views.access', args=(repo.name,))
        response = client.get(url, {'u': self.user1.username})
        self.assertEqual(response.status_code, 200)
        self.assertIn('Content-Type', response)
        self.assertEqual(response['Content-Type'], 'application/json')
        document = json.loads(response.content)
        self.assertIn('rights', document)
        self.assertEqual(document['rights'], rights)
    finally:
      settings.VCSREPO_RIGHTS_FUNCTION = original_rights_function

class SvnAuthzTestCase(BaseTestCase):
  def setUp(self):
    from ConfigParser import RawConfigParser
    super(SvnAuthzTestCase, self).setUp()
    repo = Repo(name='a', vcs='svn')
    repo.full_clean()
    repo.save()
    self.repo = repo
    self.authz = os.path.join(repo.abspath, 'conf', 'authz')
    self.config = RawConfigParser(allow_no_value=True)
    self.user1 = User.objects.create(username='user1')
    self.user2 = User.objects.create(username='user2')
    self.group1 = Group.objects.create(name='group1')
    self.group1.user_set.add(self.user1)

  def tearDown(self):
    User.objects.all().delete()
    Group.objects.all().delete()
    super(SvnAuthzTestCase, self).tearDown()

  @skipUnless(
    settings.VCSREPO_USE_USER_RIGHTS
    and settings.VCSREPO_USER_MODEL == 'auth.User',
    "not using UserRights with auth.User"
  )
  def test_add_user_rights(self):
    from .models import UserRights
    rights = UserRights.objects.create(
      repo = self.repo,
      user = self.user1,
      rights = 'r',
    )
    self.config.read(self.authz)
    self.assertTrue(self.config.has_section('/'))
    self.assertTrue(self.config.has_option('/', self.user1.username))
    self.assertEqual(self.config.get('/', self.user1.username), 'r')

  @skipUnless(
    settings.VCSREPO_USE_USER_RIGHTS
    and settings.VCSREPO_USER_MODEL == 'auth.User',
    "not using UserRights with auth.User"
  )
  def test_remove_user_rights(self):
    from .models import UserRights
    rights = UserRights.objects.create(
      repo = self.repo,
      user = self.user1,
      rights = 'r',
    )
    rights.delete()
    self.config.read(self.authz)
    self.assertTrue(self.config.has_section('/'))
    self.assertFalse(self.config.has_option('/', self.user1.username))

  @skipUnless(
    settings.VCSREPO_USE_GROUP_RIGHTS
    and settings.VCSREPO_USER_MODEL == 'auth.User'
    and settings.VCSREPO_GROUP_MODEL == 'auth.Group',
    "not using GroupRights with auth.User and auth.Group"
  )
  def test_add_group_rights(self):
    from .models import GroupRights
    rights = GroupRights.objects.create(
      repo = self.repo,
      group = self.group1,
      rights = 'r',
    )
    g = '@' + self.group1.name
    self.config.read(self.authz)
    self.assertTrue(self.config.has_section('groups'))
    self.assertTrue(self.config.has_option('groups', g))
    self.assertEqual(self.config.get('groups', g), self.user1.username)
    self.assertTrue(self.config.has_section('/'))
    self.assertTrue(self.config.has_option('/', g))
    self.assertEqual(self.config.get('/', g), 'r')

  @skipUnless(
    settings.VCSREPO_USE_GROUP_RIGHTS
    and settings.VCSREPO_USER_MODEL == 'auth.User'
    and settings.VCSREPO_GROUP_MODEL == 'auth.Group',
    "not using GroupRights with auth.User and auth.Group"
  )
  def test_remove_group_rights(self):
    from .models import GroupRights
    rights = GroupRights.objects.create(
      repo = self.repo,
      group = self.group1,
      rights = 'r',
    )
    rights.delete()
    g = '@' + self.group1.name
    self.config.read(self.authz)
    if self.config.has_section('groups'):
      self.assertFalse(self.config.has_option('groups', g))
    self.assertTrue(self.config.has_section('/'))
    self.assertFalse(self.config.has_option('/', g))

  @skipUnless(
    settings.VCSREPO_USE_GROUP_RIGHTS
    and settings.VCSREPO_USER_MODEL == 'auth.User'
    and settings.VCSREPO_GROUP_MODEL == 'auth.Group',
    "not using GroupRights with auth.User and auth.Group"
  )
  def test_add_user_to_group(self):
    from .models import GroupRights
    rights = GroupRights.objects.create(
      repo = self.repo,
      group = self.group1,
      rights = 'r',
    )
    self.group1.user_set.add(self.user2)
    g = '@' + self.group1.name
    self.config.read(self.authz)
    self.assertTrue(self.config.has_section('groups'))
    self.assertTrue(self.config.has_option('groups', g))
    members = set(map(str.strip, self.config.get('groups', g).split(',')))
    self.assertEqual(members, set([self.user1.username, self.user2.username]))
    self.assertTrue(self.config.has_section('/'))
    self.assertTrue(self.config.has_option('/', g))
    self.assertEqual(self.config.get('/', g), 'r')

  @skipUnless(
    settings.VCSREPO_USE_GROUP_RIGHTS
    and settings.VCSREPO_USER_MODEL == 'auth.User'
    and settings.VCSREPO_GROUP_MODEL == 'auth.Group',
    "not using GroupRights with auth.User and auth.Group"
  )
  def test_remove_user_from_group(self):
    from .models import GroupRights
    rights = GroupRights.objects.create(
      repo = self.repo,
      group = self.group1,
      rights = 'r',
    )
    self.group1.user_set.remove(self.user1)
    g = '@' + self.group1.name
    self.config.read(self.authz)
    self.assertTrue(self.config.has_section('groups'))
    self.assertTrue(self.config.has_option('groups', g))
    self.assertEqual(self.config.get('groups', g), '')
    self.assertTrue(self.config.has_section('/'))
    self.assertTrue(self.config.has_option('/', g))
    self.assertEqual(self.config.get('/', g), 'r')

class RepoUriTestCase(BaseTestCase):
  def test_svn(self):
    svn = Repo.objects.create(name='svn', vcs='svn', path='thesvn')
    correct = 'svn+ssh://user@hostname/svn'
    self.assertEqual(svn.ssh_uri, correct)
    svn.delete()

  def test_git(self):
    git = Repo.objects.create(name='git', vcs='git', path='thegit')
    correct = 'user@hostname:git'
    self.assertEqual(git.ssh_uri, correct)
    git.delete()

  def test_hg(self):
    hg = Repo.objects.create(name='hg', vcs='hg', path='thehg')
    correct = 'ssh://user@hostname/hg'
    self.assertEqual(hg.ssh_uri, correct)
    hg.delete()

  def test_anonymous_svn(self):
    svn = Repo.objects.create(name='svn', vcs='svn', path='thesvn')
    correct = 'svn+ssh://anonymous@hostname/svn'
    self.assertEqual(svn.anonymous_ssh_uri, correct)
    svn.delete()

  def test_anonymous_git(self):
    git = Repo.objects.create(name='git', vcs='git', path='thegit')
    correct = 'anonymous@hostname:git'
    self.assertEqual(git.anonymous_ssh_uri, correct)
    git.delete()

  def test_anonymous_hg(self):
    hg = Repo.objects.create(name='hg', vcs='hg', path='thehg')
    correct = 'ssh://anonymous@hostname/hg'
    self.assertEqual(hg.anonymous_ssh_uri, correct)
    hg.delete()

class RequestTestCase(BaseTestCase):

  def setUp(self):
    self.original_dispatch_VCSREPO_ROOT = dispatch.VCSREPO_ROOT
    dispatch.VCSREPO_ROOT = settings.VCSREPO_ROOT

  def tearDown(self):
    dispatch.VCSREPO_ROOT = self.original_dispatch_VCSREPO_ROOT

  def test_git_cmd1(self):
    request = dispatch.get_request(['git-upload-pack', 'bob/code'])
    request.data = {'rights': 'r', 'path': 'path/to/code'}
    cmd = request.get_command()
    expected = ['git', 'shell', '-c',
                "git-upload-pack 'path/to/code'"]
    self.assertEqual(expected, cmd)

  def test_git_cmd2(self):
    request = dispatch.get_request(['git-upload-pack', 'bob/code'])
    request.data = {'rights': '', 'path': 'path/to/code'}
    self.assertRaises(dispatch.DispatchException, request.get_command)

  def test_git_cmd3(self):
    request = dispatch.get_request(['git-receive-pack', 'bob/code'])
    request.data = {'rights': 'r', 'path': 'path/to/code'}
    self.assertRaises(dispatch.DispatchException, request.get_command)

  def test_git_cmd4(self):
    request = dispatch.get_request(['git-upload-pack', '--someflag', 'bob/code'])
    request.data = {'rights': 'r', 'path': 'path/to/code'}
    cmd = request.get_command()
    expected = ['git', 'shell', '-c',
                "git-upload-pack 'path/to/code'"]
    self.assertEqual(expected, cmd)

  def test_git_repo_name1(self):
    request = dispatch.get_request(['git-upload-pack', 'bob/code'])
    self.assertEqual('bob/code', request.repo_name)

  def test_git_repo_name1(self):
    request = dispatch.get_request(['git-upload-pack', 'bob/code.git'])
    self.assertEqual('bob/code', request.repo_name)

  def test_git_noname(self):
    self.assertRaises(dispatch.DispatchException, dispatch.get_request, ['git-receive-pack'])

  def test_hg_cmd1(self):
    '''Repository specified with -R
    '''
    request = dispatch.get_request(['hg', '-R', 'bob/code'])
    request.data = {'rights': 'rw', 'path': 'path/to/code'}
    cmd = request.get_command()
    expected = ['hg', '-R', 'path/to/code', 'serve', '--stdio']
    self.assertEqual(expected, cmd)

  def test_hg_cmd2(self):
    '''Repository specified with --repository
    '''
    request = dispatch.get_request(['hg', '--repository', 'bob/code'])
    request.data = {'rights': 'rw', 'path': 'path/to/code'}
    cmd = request.get_command()
    expected = ['hg', '-R', 'path/to/code', 'serve', '--stdio']
    self.assertEqual(expected, cmd)

  def test_hg_cmd3(self):
    '''Read-only access.
    '''
    request = dispatch.get_request(['hg', '--repository', 'bob/code'])
    request.data = {'rights': 'r', 'path': 'path/to/code'}
    cmd = request.get_command()
    expected = ['hg', '-R', 'path/to/code', 'serve', '--stdio',
        '--config',
        'hooks.prechangegroup.readonly=echo "Error: Permission denied (read-only)" >&2; false',
        '--config',
        'hooks.prepushkey.readonly=echo "Error: Permission denied (read-only)" >&2; false',
    ]
    self.assertEqual(expected, cmd)

  def test_svn_cmd1(self):
    '''Without username.
    '''
    request = dispatch.get_request(['svnserve', '--root', 'bob/code'])
    request.data = {'rights': 'r', 'path': 'path/to/code'}
    cmd = request.get_command()
    expected = ['svnserve',
                '--root',  os.path.join(settings.VCSREPO_ROOT, 'svn'),
                '--tunnel']
    self.assertEqual(expected, cmd)

  def test_svn_cmd2(self):
    '''With username.
    '''
    request = dispatch.get_request(['svnserve', '--root', 'bob/code'], 'bob')
    request.data = {'rights': 'r', 'path': 'path/to/code'}
    cmd = request.get_command()
    expected = ['svnserve',
                '--root',  os.path.join(settings.VCSREPO_ROOT, 'svn'),
                '--tunnel',
                '--tunnel-user', 'bob']
    self.assertEqual(expected, cmd)

  def test_bad_command1(self):
    self.assertRaises(dispatch.DispatchException, dispatch.get_request, ['rm', '-rf', '/'])

  def test_hg_postprocess(self):
    request = dispatch.get_request(['hg', '--repository', 'bob/code'])
    request.data = {'rights': 'r', 'path': 'path/to/code'}
    result = request.postprocess('hg: cloning from path/to/code')
    expected = 'hg: cloning from bob/code'
    self.assertEqual(expected, result)

class PristineTestCase(BaseTestCase):
  '''
  Normal, pristine repository.
  '''

  def setUp(self):
    super(PristineTestCase, self).setUp()
    self.repo = Repo(name='repo', vcs='git')
    self.repo.full_clean()
    self.repo.save()

  def test_get_entry_or_404_fail1(self):
    self.assertRaises(Http404, shortcuts.get_entry_or_404,
                      self.repo, 'notexist', 'notexist')

class NormalContentsTestCase(BaseTestCase):
  '''
  Repository with some contents for testing purposes.

  * rev1: One of every file type, including subdirs.
  * rev2: Empty tree.
  * rev3: Text and binary files.

  Structure at rev1:
  .
  ├── a
  ├── b
  │   └── c
  └── d -> a

  Structure at rev3:
  .
  ├── encoding.txt
  ├── text.txt
  └── binary.gz

  '''

  def setUp(self):
    super(NormalContentsTestCase, self).setUp()
    self.repo = Repo(name='repo', vcs='git')
    self.repo.full_clean()
    self.repo.save()

    # Basic repository setup.
    wc = tempfile.mktemp()
    cmd = [GIT, 'clone', '-q', self.repo.abspath, wc]
    subprocess.check_call(cmd, stderr=DEVNULL)
    setup_git(cwd=wc)

    # rev1 setup
    with open(os.path.join(wc, 'a'), 'w') as fp: pass
    os.makedirs(os.path.join(wc, 'b'))
    with open(os.path.join(wc, 'b', 'c'), 'w') as fp: pass
    os.symlink('a', os.path.join(wc, 'd'))
    cmd = [GIT, 'add', '-A', '.']
    subprocess.check_call(cmd, cwd=wc)
    cmd = [GIT, 'commit', '-q', '-m', 'initial commit']
    subprocess.check_call(cmd, cwd=wc)

    # rev2 setup
    cmd = [GIT, 'rm', '-rq', '.']
    subprocess.check_call(cmd, cwd=wc)
    cmd = [GIT, 'commit', '-q', '-m', 'remove all files']
    subprocess.check_call(cmd, cwd=wc)

    # rev3 setup
    with open(os.path.join(wc, 'text.txt'), 'w') as fp:
      fp.write('hello\n')
    with open(os.path.join(wc, 'encoding.txt'), 'w') as fp:
      fp.write(u'P\xe9rez\n'.encode('latin1'))
    with gzip.open(os.path.join(wc, 'binary.gz'), 'wb') as fp:
      fp.write('hello\n')
    cmd = [GIT, 'add', '-A', '.']
    subprocess.check_call(cmd, cwd=wc)
    cmd = [GIT, 'commit', '-q', '-m', 'text and binary file']
    subprocess.check_call(cmd, cwd=wc)

    # Push the result.
    cmd = [GIT, 'push', '-q', '-u', 'origin', 'master']
    subprocess.check_call(cmd, cwd=wc, stdout=DEVNULL)

    # Set up some easy names.
    self.branch = 'master'
    self.rev1 = self.repo.repo.canonical_rev(self.branch + '~2')
    self.rev2 = self.repo.repo.canonical_rev(self.branch + '~1')
    self.rev3 = self.repo.repo.canonical_rev(self.branch + '~0')
    shutil.rmtree(wc)
    self.repo = Repo.objects.get()

  def test_get_entry_or_404_file1(self):
    '''Standard file test'''
    entry = shortcuts.get_entry_or_404(self.repo, self.rev1, '/a')
    self.assertEqual('a', entry.path)
    self.assertEqual('f', entry.type)

  def test_get_entry_or_404_file2(self):
    '''File in directory structure test'''
    entry = shortcuts.get_entry_or_404(self.repo, self.rev1, '/b/c')
    self.assertEqual('b/c', entry.path)
    self.assertEqual('f', entry.type)

  def test_get_entry_or_404_dir1(self):
    '''Standard directory test'''
    entry = shortcuts.get_entry_or_404(self.repo, self.rev1, '/b')
    self.assertEqual('b', entry.path)
    self.assertEqual('d', entry.type)

  def test_get_entry_or_404_dir2(self):
    '''Listing contents of empty tree should not 404'''
    entry = shortcuts.get_entry_or_404(self.repo, self.rev2, '/')
    self.assertEqual('/', entry.path)
    self.assertEqual('d', entry.type)

  def test_get_entry_or_404_link1(self):
    '''Standard link test'''
    entry = shortcuts.get_entry_or_404(self.repo, self.rev1, '/d')
    self.assertEqual('d', entry.path)
    self.assertEqual('l', entry.type)

  def test_get_entry_or_404_report1(self):
    '''Keyword arguments should pass through'''
    entry = shortcuts.get_entry_or_404(self.repo, self.rev1, '/a',
                                       report=('commit',))
    self.assertEqual('a', entry.path)
    self.assertEqual('f', entry.type)
    self.assertEqual(self.rev1, entry.commit)

  def test_get_entry_or_404_fail1(self):
    '''Bad paths raise 404'''
    self.assertRaises(Http404, shortcuts.get_entry_or_404,
                      self.repo, self.rev1, 'notexist')

  def test_get_entry_or_404_fail2(self):
    '''Bad revisions raise 404'''
    self.assertRaises(Http404, shortcuts.get_entry_or_404,
                      self.repo, 'notexist', '/a')

  def test_get_entry_or_404_fail3(self):
    '''Extra bad path test for the empty tree case'''
    self.assertRaises(Http404, shortcuts.get_entry_or_404,
                      self.repo, self.rev2, 'a')

  def test_get_directory_contents1(self):
    '''Basic usage'''
    result = shortcuts.get_directory_contents(self.repo, self.rev1, '/')
    expected = [
      {'name': 'a', 'path': 'a', 'type': 'f', 'url': 'a'},
      {'name': 'b', 'path': 'b', 'type': 'd', 'url': 'b'},
      {'name': 'd', 'path': 'd', 'type': 'l'},
    ]
    self.assertEqual(result, expected)

  def test_get_directory_contents2(self):
    '''Keyword arguments are passed through'''
    result = shortcuts.get_directory_contents(self.repo, self.rev1, '/',
                                              report=('target',))
    expected = [
      {'name': 'a', 'path': 'a', 'type': 'f', 'url': 'a'},
      {'name': 'b', 'path': 'b', 'type': 'd', 'url': 'b'},
      {'name': 'd', 'path': 'd', 'type': 'l', 'target': 'a'},
    ]
    self.assertEqual(result, expected)

  def test_get_directory_contents3(self):
    '''Test reverse sort parameter'''
    result = shortcuts.get_directory_contents(self.repo, self.rev1, '/',
                                              reverse=True)
    expected = [
      {'name': 'd', 'path': 'd', 'type': 'l'},
      {'name': 'b', 'path': 'b', 'type': 'd', 'url': 'b'},
      {'name': 'a', 'path': 'a', 'type': 'f', 'url': 'a'},
    ]
    self.assertEqual(result, expected)

  def test_get_directory_contents4(self):
    '''Test sort key parameter'''
    result = shortcuts.get_directory_contents(self.repo, self.rev1, '/',
                                              key=lambda e: e.type)
    expected = [
      {'name': 'b', 'path': 'b', 'type': 'd', 'url': 'b'},
      {'name': 'a', 'path': 'a', 'type': 'f', 'url': 'a'},
      {'name': 'd', 'path': 'd', 'type': 'l'},
    ]
    self.assertEqual(result, expected)

  def test_get_directory_contents5(self):
    '''Test resolve_commits'''
    result = [e.log.rev for e in
              shortcuts.get_directory_contents(self.repo, self.rev1, '/',
                                               resolve_commits=True)]
    log = self.repo.repo.log(revrange=self.rev1)
    expected = [self.rev1] * 3
    self.assertEqual(result, expected)

  def test_get_directory_contents_subdir1(self):
    '''Basic usage'''
    result = shortcuts.get_directory_contents(self.repo, self.rev1, '/b')
    expected = [
      {'name': '..', 'path': '', 'type': 'd', 'url': '..'},
      {'name': 'c', 'path': 'b/c', 'type': 'f', 'url': 'c'},
    ]
    self.assertEqual(result, expected)

  def test_get_directory_contents_subdir2(self):
    '''Disallow parent paths'''
    result = shortcuts.get_directory_contents(self.repo, self.rev1, '/b',
                                              parents=False)
    expected = [
      {'name': 'c', 'path': 'b/c', 'type': 'f', 'url': 'c'},
    ]
    self.assertEqual(result, expected)

  def test_get_directory_contents_subdir3(self):
    '''Using a custom reverse_func'''
    reverse_func = lambda e: 'http://example.com/repo/' + e.path
    result = shortcuts.get_directory_contents(self.repo, self.rev1, '/b',
                                              reverse_func=reverse_func)
    expected = [
      {'name': '..', 'path': '', 'type': 'd', 'url': 'http://example.com/repo/'},
      {'name': 'c', 'path': 'b/c', 'type': 'f', 'url': 'http://example.com/repo/b/c'},
    ]
    self.assertEqual(result, expected)

  def test_render_file1(self):
    '''Basic text gets rendered into the template'''
    result = shortcuts.render_file('raw.html',
                                   self.repo, self.rev3, '/text.txt')
    self.assertEqual('text/html; charset=utf-8', result['Content-Type'])

  def test_render_file2(self):
    '''Raw argument adds mimetype and returns the file.'''
    result = shortcuts.render_file('raw.html',
                                   self.repo, self.rev3, '/text.txt',
                                   raw=True)
    self.assertEqual('text/plain', result['Content-Type'])

  def test_render_file3(self):
    '''Test raw flag with mimetype override'''
    result = shortcuts.render_file('raw.html',
                                   self.repo, self.rev3, '/text.txt',
                                   file_mimetype='text/x-csrc',
                                   raw=True)
    self.assertEqual('text/x-csrc', result['Content-Type'])

  def test_render_file4(self):
    '''Non-text files get passed through raw.'''
    result = shortcuts.render_file('raw.html',
                                   self.repo, self.rev3, '/binary.gz')
    self.assertEqual('application/octet-stream', result['Content-Type'])

  def test_render_file5(self):
    '''Encoding errors are properly raised.'''
    with self.assertRaises(DjangoUnicodeDecodeError):
      shortcuts.render_file('raw.html',
                            self.repo, self.rev3, '/encoding.txt')

  def test_render_file6(self):
    '''The `catch_endcoding_errors` argument handles encoding errors.'''
    result = shortcuts.render_file('raw.html',
                                   self.repo, self.rev3, '/encoding.txt',
                                   catch_encoding_errors=True)
    self.assertEqual('text/plain', result['Content-Type'])

  def test_render_file7(self):
    '''Test the textfilter parameter'''
    def textfilter(contents, path, mimetype):
      self.assertIsInstance(contents, basestring)
      self.assertEqual('/text.txt', path)
      self.assertEqual('text/plain', mimetype)
      return 'test - ' + contents
    response = shortcuts.render_file('raw.html',
                                     self.repo, self.rev3, '/text.txt',
                                     textfilter=textfilter)
    self.assertEqual('test - hello\n\n', response.content)

  def test_render_file8(self):
    '''Test the context'''
    response = self.client.get('/browse/repo/%s/text.txt' % self.rev3)
    self.assertEqual(self.repo.pk, response.context['repo'].pk)
    self.assertEqual('/text.txt', response.context['path'])
    self.assertEqual(self.rev3, response.context['rev'])
    self.assertIsInstance(response.context['contents'], basestring)


def setup_git(**kw):
  cmd = [GIT, 'config', 'user.name', 'Test User']
  subprocess.check_call(cmd, **kw)
  cmd = [GIT, 'config', 'user.email', 'test@example.com']
  subprocess.check_call(cmd, **kw)



# Copyright (c) 2014-2015, Clemson University
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of Clemson University nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from django.conf import settings
from . import defaults
import getpass
import socket

VCSREPO_ROOT = settings.VCSREPO_ROOT
VCSREPO_PATH_FUNCTION = getattr(settings, 'VCSREPO_RELPATH_FUNCTION', defaults.path_function)
VCSREPO_CHECK_NESTED_PATHS = getattr(settings, 'VCSREPO_CHECK_NESTED_PATHS', True)
VCSREPO_ALLOW_NESTED_PATHS = getattr(settings, 'VCSREPO_ALLOW_NESTED_PATHS', False)

VCSREPO_USE_USER_RIGHTS = getattr(settings, 'VCSREPO_USE_USER_RIGHTS', True)
VCSREPO_USE_GROUP_RIGHTS = getattr(settings, 'VCSREPO_USE_GROUP_RIGHTS', True)

VCSREPO_USER_MODEL = getattr(settings, 'VCSREPO_USER_MODEL', getattr(settings, 'AUTH_USER_MODEL', 'auth.User'))
VCSREPO_GROUP_MODEL = getattr(settings, 'VCSREPO_GROUP_MODEL', 'auth.Group')

VCSREPO_RIGHTS_FUNCTION = getattr(settings, 'VCSREPO_RIGHTS_FUNCTION', defaults.rights_function)
VCSREPO_USER_ACL_FUNCTION = getattr(settings, 'VCSREPO_USER_ACL_FUNCTION', defaults.user_acl_function if VCSREPO_USE_USER_RIGHTS else None)
VCSREPO_GROUP_ACL_FUNCTION = getattr(settings, 'VCSREPO_GROUP_ACL_FUNCTION', defaults.group_acl_function if VCSREPO_USE_GROUP_RIGHTS else None)

if hasattr(settings, 'VCS_URI_FORMAT'):
  import warnings
  warn = "VCS_URI_FORMAT is deprecated; use VCSREPO_URI_FORMAT"
  warnings.warn(warn, DeprecationWarning)
  VCSREPO_URI_FORMAT = settings.VCS_URI_FORMAT
else:
  VCSREPO_URI_FORMAT = getattr(settings, 'VCSREPO_URI_FORMAT', {})
VCSREPO_URI_FORMAT.setdefault(('git', 'ssh'), '{user}@{hostname}:{path}')
VCSREPO_URI_FORMAT.setdefault(('svn', 'ssh'), 'svn+ssh://{user}@{hostname}/{path}')
VCSREPO_URI_FORMAT.setdefault(('hg', 'ssh'), 'ssh://{user}@{hostname}/{path}')
VCSREPO_URI_FORMAT.setdefault(('git', 'anonymous-ssh'), '{anonymous}@{hostname}:{path}')
VCSREPO_URI_FORMAT.setdefault(('svn', 'anonymous-ssh'), 'svn+ssh://{anonymous}@{hostname}/{path}')
VCSREPO_URI_FORMAT.setdefault(('hg', 'anonymous-ssh'), 'ssh://{anonymous}@{hostname}/{path}')

if hasattr(settings, 'VCS_URI_CONTEXT'):
  import warnings
  warn = "VCS_URI_CONTEXT is deprecated; use VCSREPO_URI_CONTEXT"
  warnings.warn(warn, DeprecationWarning)
  VCSREPO_URI_CONTEXT = settings.VCS_URI_CONTEXT
else:
  VCSREPO_URI_CONTEXT = getattr(settings, 'VCSREPO_URI_CONTEXT', {})
VCSREPO_URI_CONTEXT.setdefault('anonymous', 'anonymous')
VCSREPO_URI_CONTEXT.setdefault('user', getpass.getuser())
VCSREPO_URI_CONTEXT.setdefault('hostname', socket.gethostname())



# Copyright (c) 2014-2015, Clemson University
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of Clemson University nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import logging
import os
import sys

logger = logging.getLogger(__name__)

VCSREPO_ROOT = os.getenv('VCSREPO_ROOT')
GIT = os.getenv('GIT', 'git')
HG = os.getenv('HG', 'hg')
SVNSERVE = os.getenv('SVNSERVE', 'svnserve')

class DispatchException(Exception):
  pass

class Request(object):
  postprocess = None

  def __init__(self, argv, username):
    self.argv = argv
    self.username = username
    self.repo_name = None
    self.data = {}

  def add_data(self, data):
    self.data.update(data)

  def load_data(self, url, params=None):
    import json
    import urllib
    if params:
      url += '?' + urllib.urlencode(params)
    response = urllib.urlopen(url)
    status = response.getcode()
    content_type = response.info().gettype()
    if status == 200:
      data = json.load(response)
      self.add_data(data)
    else:
      if content_type == 'text/plain':
        raise DispatchException(response.readline().strip())
      if content_type == 'application/json':
        data = json.load(response)
        if isinstance(data, dict) and 'error' in data:
          raise DispatchException(data['error'])
        if isinstance(data, basestring):
          raise DispatchException(data)
      raise DispatchException('Backend failed', status)

  def get_command(self):
    raise NotImplementedError

  def run_command(self):
    import subprocess
    cmd = self.get_command()
    if not self.postprocess:
      return subprocess.call(cmd)
    p = subprocess.Popen(cmd, stderr=subprocess.PIPE)
    stdout, stderr = p.communicate()
    stderr = self.postprocess(stderr)
    sys.stderr.write(stderr)
    return p.returncode

class GitRequest(Request):
  vcs = 'git'

  def __init__(self, argv, username):
    super(GitRequest, self).__init__(argv, username)
    for arg in argv[1:]:
      if not arg.startswith('-'):
        repo_name = arg
        break
    else:
      raise DispatchException('Repository not specified')
    repo_name = repo_name.lstrip('/')
    if repo_name.endswith('.git'):
      repo_name = repo_name[:-4]
    self.repo_name = repo_name

  def get_command(self):
    rights = self.data['rights']
    path = self.data['path']
    if 'r' not in rights:
      raise DispatchException('Permission denied')
    if self.argv[0] == 'git-receive-pack':
      if 'w' not in rights:
        raise DispatchException('Permission denied (read only)')
    cmd = [GIT, 'shell', '-c', "%s '%s'" % (self.argv[0], path)]
    return cmd

class HgRequest(Request):
  vcs = 'hg'

  def __init__(self, argv, username):
    super(HgRequest, self).__init__(argv, username)
    for i in range(1, len(argv)-1):
      if argv[i] in ('-R', '--repository'):
        repo_name = argv[i+1]
        break
    else:
      raise DispatchException('Repository not specified')
    self.repo_name = repo_name.lstrip('/')

  def get_command(self):
    rights = self.data['rights']
    path = self.data['path']
    if 'r' not in rights:
      raise DispatchException('Permission denied')
    cmd = [HG, '-R', path, 'serve', '--stdio']
    if 'w' not in rights:
      cmd += [
        '--config',
        'hooks.prechangegroup.readonly=echo "Error: Permission denied (read-only)" >&2; false',
        '--config',
        'hooks.prepushkey.readonly=echo "Error: Permission denied (read-only)" >&2; false',
      ]
    return cmd

  def postprocess(self, text):
    return text.replace(self.data['path'], self.repo_name)

class SvnRequest(Request):
  vcs = 'svn'

  def get_command(self):
    assert VCSREPO_ROOT is not None, 'VCSREPO_ROOT is not set'
    svn_dir = os.path.join(VCSREPO_ROOT, 'svn')
    cmd = [SVNSERVE, '--root', svn_dir, '--tunnel']
    if self.username is not None:
      cmd.extend(['--tunnel-user', self.username])
    return cmd


def parse_command(cmd):
  import shlex
  try:
    argv = shlex.split(cmd)
  except ValueError as e:
    raise DispatchException('Illegal command', e)
  if not argv:
    raise DispatchException('Command not specified')
  return argv

def get_request(argv, username=None):
  if argv[0] in ('git-receive-pack', 'git-upload-pack', 'git-upload-archive'):
    return GitRequest(argv, username)
  if argv[0] == 'hg':
    return HgRequest(argv, username)
  if argv[0] == 'svnserve':
    return SvnRequest(argv, username)
  raise DispatchException('Command not allowed', argv[0])

def ssh_dispatch(access_url, username):
  cmd = os.getenv('SSH_ORIGINAL_COMMAND', '')
  try:
    argv = parse_command(cmd)
    request = get_request(argv, username)
    if request.repo_name:
      url = '%s/%s' % (access_url, request.repo_name)
      params = {'vcs': request.vcs}
      if username:
        params['u'] = username
      request.load_data(url, params)
    return request.run_command()
  except DispatchException as e:
    sys.stderr.write('Error: ' + str(e) + '\n')
    sys.exit(1)
  except:
    message = 'Unhandled error during command dispatch'
    logger.error(message, exc_info=True)
    sys.stderr.write('%s\n' % message)
    sys.exit(1)

def main():
  argc = len(sys.argv)
  if argc < 2 or argc > 3:
    sys.stderr.write('Usage: %s <access-url> [<username>]\n' % sys.argv[0])
    sys.exit(1)
  url = sys.argv[1]
  username = None
  if argc == 3:
    username = sys.argv[2]

  status = ssh_dispatch(url, username)
  sys.exit(status)

if __name__ == '__main__':
  main()



# Copyright (c) 2014-2015, Clemson University
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of Clemson University nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from django.contrib import admin
from . import settings
from .models import Repo

def update_svnserve(modeladmin, request, queryset):
  for obj in queryset:
    obj.update_svnserve()
update_svnserve.short_description = 'Update svnserve.conf'

def relocate_path(modeladmin, request, queryset):
  for obj in queryset:
    obj.relocate_path()
    obj.save()

class RepoAdmin(admin.ModelAdmin):
  list_display = ['__unicode__', 'path', 'vcs']
  list_filter = ['vcs']
  search_fields = ['name', 'path']
  actions = [
    update_svnserve,
    relocate_path,
  ]

  def get_readonly_fields(self, request, obj=None):
    if obj:
      return ['abspath', 'vcs']
    else:
      return []

  def abspath(self, instance):
    return instance.abspath
  abspath.short_description = 'Full Path'

admin.site.register(Repo, RepoAdmin)

if settings.VCSREPO_USE_USER_RIGHTS:
  from .models import UserRights

  class UserRightsAdmin(admin.ModelAdmin):
    list_display = ['__unicode__', 'repo', 'user', 'rights']
    search_fields = ['repo__name', 'user__username']

  admin.site.register(UserRights, UserRightsAdmin)

if settings.VCSREPO_USE_GROUP_RIGHTS:
  from .models import GroupRights

  class GroupRightsAdmin(admin.ModelAdmin):
    list_display = ['__unicode__', 'repo', 'group', 'rights']
    search_fields = ['repo__name', 'group__name']

  admin.site.register(GroupRights, GroupRightsAdmin)



# Copyright (c) 2014-2015, Clemson University
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of Clemson University nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

def path_function(repo):
  import hashlib
  import os
  h = hashlib.sha1(repo.name).hexdigest()
  p = [
    repo.vcs,
    h[0:2],
    h[2:4],
    h[4:],
  ]
  return os.path.join(*p)

def rights_function(repo, user):
  from . import settings
  if user is not None:
    if settings.VCSREPO_USE_USER_RIGHTS:
      from .models import UserRights
      try:
        userrights = UserRights.objects.get(repo=repo, user=user)
        return userrights.rights
      except UserRights.DoesNotExist:
        pass
    if settings.VCSREPO_USE_GROUP_RIGHTS:
      from .models import GroupRights
      rights = None
      for group in user.groups.all():
        try:
          grouprights = GroupRights.objects.get(repo=repo, group=group)
          if rights is None or len(grouprights.rights) < rights:
            rights = grouprights.rights
        except GroupRights.DoesNotExist:
          pass
      if rights is not None:
        return rights
  if repo.public_read:
    return 'r'
  return '-'

def user_acl_function(repo):
  return dict((x.user, x.rights) for x in repo.userrights_set.all())

def group_acl_function(repo):
  return dict((x.group, x.rights) for x in repo.grouprights_set.all())



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations
from django.conf import settings


class Migration(migrations.Migration):

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name='GroupRights',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('rights', models.CharField(default=b'rw', max_length=2, choices=[(b'-', b'Deny'), (b'r', b'Read-Only'), (b'rw', b'Read-Write')])),
                ('created', models.DateTimeField(auto_now_add=True, null=True)),
                ('last_modified', models.DateTimeField(auto_now=True, null=True)),
                ('group', models.ForeignKey(to='auth.Group')),
            ],
            options={
                'db_table': 'anyvcs_grouprights',
                'verbose_name': 'Group Access Rights',
                'verbose_name_plural': 'Group Access Rights',
            },
        ),
        migrations.CreateModel(
            name='Repo',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('name', models.CharField(unique=True, max_length=100, db_index=True)),
                ('path', models.CharField(help_text=b'Either relative to VCSREPO_ROOT or absolute. Changing this will move the repository on disk.', unique=True, max_length=100, blank=True)),
                ('vcs', models.CharField(default=b'git', max_length=3, verbose_name=b'Version Control System', choices=[(b'git', b'Git'), (b'hg', b'Mercurial'), (b'svn', b'Subversion')])),
                ('public_read', models.BooleanField(default=False, verbose_name=b'Public Read Access')),
                ('created', models.DateTimeField(auto_now_add=True, null=True)),
                ('last_modified', models.DateTimeField(auto_now=True, null=True)),
            ],
            options={
                'ordering': ['name'],
                'db_table': 'anyvcs_repo',
                'verbose_name': 'Repository',
                'verbose_name_plural': 'Repositories',
            },
        ),
        migrations.CreateModel(
            name='UserRights',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('rights', models.CharField(default=b'rw', max_length=2, choices=[(b'-', b'Deny'), (b'r', b'Read-Only'), (b'rw', b'Read-Write')])),
                ('created', models.DateTimeField(auto_now_add=True, null=True)),
                ('last_modified', models.DateTimeField(auto_now=True, null=True)),
                ('repo', models.ForeignKey(to='django_anyvcs.Repo')),
                ('user', models.ForeignKey(to=settings.AUTH_USER_MODEL)),
            ],
            options={
                'db_table': 'anyvcs_userrights',
                'verbose_name': 'User Access Rights',
                'verbose_name_plural': 'User Access Rights',
            },
        ),
        migrations.AddField(
            model_name='grouprights',
            name='repo',
            field=models.ForeignKey(to='django_anyvcs.Repo'),
        ),
        migrations.AlterUniqueTogether(
            name='userrights',
            unique_together=set([('repo', 'user')]),
        ),
        migrations.AlterUniqueTogether(
            name='grouprights',
            unique_together=set([('repo', 'group')]),
        ),
    ]



# Copyright (c) 2014-2015, Clemson University
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of Clemson University nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from anyvcs.common import PathDoesNotExist, attrdict
from django.http import Http404, HttpResponse
from django.shortcuts import render_to_response
from django.utils.encoding import force_text

import mimetypes
import os

def get_entry_or_404(repo, rev, path, **kw):
  '''
  Get entry via `repo.repo.ls()` or raise Http404.

  Keyword arguments are passed along to the call to `ls()`.

  '''
  if not rev in repo.repo:
    raise Http404
  try:
    return repo.repo.ls(rev, path, directory=True, **kw)[0]
  except PathDoesNotExist:
    raise Http404

def get_directory_contents(repo, rev, path, key=None, reverse=False,
                           parents=True, reverse_func=None,
                           resolve_commits=False, **kw):
  '''
  Get repository contents suitable for using in a template context.

  With `key` and `reverse`, modify the sort order via `sorted()` of the
  directory contents. The default is to sort by name.

  With `parents` modify whether or not to include links to parent directories.
  The default is to include them.

  Provide a callable `reverse_func` to transform an entry into a URL. The
  default is to generate relative links.

  With `resolve_commits`, also include the commit log entry which last modified
  the entry. The default is to not. This forces commits to be reported from the
  call to `VCSRepo.ls()`.

  '''
  path = _normpath(path)
  key = key or (lambda e: e.name)

  # Force the report commit flag if not specified and resolve_commits is True.
  report = tuple(kw.get('report', ()))
  if resolve_commits and not 'commit' in report:
    report += ('commit',)
    kw['report'] = report

  contents = sorted(repo.repo.ls(rev, path, **kw), key=key, reverse=reverse)

  # Use relative paths by default for the URL.
  reverse_func = reverse_func or (lambda e: e.name)

  # Loop over the contents and add extra information in.
  cache = {}
  for entry in contents:
    if entry.type != 'l':
      entry.url = reverse_func(entry)
    # XXX The check for the commit is a hack around python-anyvcs#65
    if resolve_commits and 'commit' in entry:
      commit = entry.commit
      try:
        entry.log = cache[commit]
      except KeyError:
        entry.log = cache[commit] = repo.repo.log(revrange=commit)

  # Add parent directories if requested.
  parent_path = _normpath(_pardir(path))
  if path != '/' and parents:
    entry = attrdict(name='..', path=parent_path.lstrip('/'), type='d')
    entry.url = reverse_func(entry)
    contents.insert(0, entry)
  return contents

def render_file(template, repo, rev, path, file_mimetype=None,
                encoding='utf-8', extra_context=None, textfilter=None,
                raw=False, contents=None, catch_encoding_errors=False, **kw):
  '''
  Render a file to a template, or return the raw contents of the file if it is
  not a text file.

  With `file_mimetype` you can modify the mimetype of the file. The default is
  to detect the mimetype via the path and `mimetypes.guess_type()`. The
  default `encoding` can also be changed. You should catch the encoding errors
  which correspond to your chosen codec.

  `extra_context` may be a dict-like object which are passed into context for
  the template. By default, only `contents`, `rev`, and `path` are passed into
  the template.

  With `textfilter`, you can provide a filter function which modifies the
  decoded text contents of the file. The function must take the file contents,
  the path, and the file's mimetype as its three positional arguments. The
  function is only called if the file is detected to be a text file.

  If `raw` is True, the file will unconditionally be returned in its original
  form.

  If `contents` is given, the contents of the file are not retrieved, and it is
  assumed that `contents` contains the contents of the file.

  If `catch_encoding_errors` is True, encoding errors are handled by returning
  the raw file. Otherwise, you must handle the encoding errors.

  '''
  if contents is None:
    contents = repo.repo.cat(rev, path)

  if file_mimetype is None:
    file_mimetype, _ = mimetypes.guess_type(path)
    file_mimetype = file_mimetype or 'application/octet-stream'

  if raw or not file_mimetype.startswith('text/'):
    return HttpResponse(contents, content_type=file_mimetype)

  try:
    decoded_contents = force_text(contents, encoding=encoding)
  except:
    if not catch_encoding_errors:
      raise
    return HttpResponse(contents, content_type=file_mimetype)

  if textfilter is not None:
    decoded_contents = textfilter(decoded_contents, path, file_mimetype)

  context = {
    'repo': repo,
    'contents': decoded_contents,
    'rev': rev,
    'path': path,
  }
  if extra_context is not None:
    context.update(extra_context)

  return render_to_response(template, context, **kw)

def _normpath(path):
  path = os.path.normpath(path)
  if path in ('.', '/'):
    return '/'
  return path

def _pardir(path, levels=1):
  args = ['..'] * levels
  return os.path.join(path, *args)



# Copyright (c) 2014-2015, Clemson University
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of Clemson University nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import os
from setuptools import setup

README = open(os.path.join(os.path.dirname(__file__), 'README.rst')).read()

# allow setup.py to be run from any path
os.chdir(os.path.normpath(os.path.join(os.path.abspath(__file__), os.pardir)))

exec(open('django_anyvcs/__init__.py').read())

setup(
  name='django-anyvcs',
  version=__version__,
  packages=['django_anyvcs'],
  include_package_data=True,
  license='BSD',
  description='A Django app providing homogeneous management of VCS systems.',
  long_description=README,
  url='https://github.com/ClemsonSoCUnix/django-anyvcs',
  author='Scott Duckworth',
  author_email='sduckwo@clemson.edu',
  zip_safe=False,
  classifiers=[
    'Environment :: Web Environment',
    'Framework :: Django',
    'Intended Audience :: Developers',
    'License :: OSI Approved :: BSD License',
    'Operating System :: OS Independent',
    'Programming Language :: Python',
    'Programming Language :: Python :: 2.6',
    'Programming Language :: Python :: 2.7',
    'Topic :: Internet :: WWW/HTTP',
    'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
  ],
  install_requires=[
    'anyvcs',
  ],
  entry_points = {
    'console_scripts': [
      'django-anyvcs-ssh = django_anyvcs.dispatch:main',
    ],
  },
)



# encoding: utf-8
import datetime
from south.db import db
from south.v2 import DataMigration
from django.db import models

class Migration(DataMigration):

  def forwards(self, orm):
    for repo in orm.Repo.objects.all():
      repo.public_read = 'r' in repo.public_rights
      repo.save()

  def backwards(self, orm):
    for repo in orm.Repo.objects.all():
      # it won't be possible to recover public write
      if repo.public_read:
        repo.public_rights = 'r'
      else:
        repo.public_rights = '-'
      repo.save()


  models = {
    'auth.group': {
      'Meta': {'object_name': 'Group'},
      'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
      'name': ('django.db.models.fields.CharField', [], {'unique': 'True', 'max_length': '80'}),
      'permissions': ('django.db.models.fields.related.ManyToManyField', [], {'to': "orm['auth.Permission']", 'symmetrical': 'False', 'blank': 'True'})
    },
    'auth.permission': {
      'Meta': {'ordering': "('content_type__app_label', 'content_type__model', 'codename')", 'unique_together': "(('content_type', 'codename'),)", 'object_name': 'Permission'},
      'codename': ('django.db.models.fields.CharField', [], {'max_length': '100'}),
      'content_type': ('django.db.models.fields.related.ForeignKey', [], {'to': "orm['contenttypes.ContentType']"}),
      'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
      'name': ('django.db.models.fields.CharField', [], {'max_length': '50'})
    },
    'auth.user': {
      'Meta': {'object_name': 'User'},
      'date_joined': ('django.db.models.fields.DateTimeField', [], {'default': 'datetime.datetime.now'}),
      'email': ('django.db.models.fields.EmailField', [], {'max_length': '75', 'blank': 'True'}),
      'first_name': ('django.db.models.fields.CharField', [], {'max_length': '30', 'blank': 'True'}),
      'groups': ('django.db.models.fields.related.ManyToManyField', [], {'to': "orm['auth.Group']", 'symmetrical': 'False', 'blank': 'True'}),
      'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
      'is_active': ('django.db.models.fields.BooleanField', [], {'default': 'True'}),
      'is_staff': ('django.db.models.fields.BooleanField', [], {'default': 'False'}),
      'is_superuser': ('django.db.models.fields.BooleanField', [], {'default': 'False'}),
      'last_login': ('django.db.models.fields.DateTimeField', [], {'default': 'datetime.datetime.now'}),
      'last_name': ('django.db.models.fields.CharField', [], {'max_length': '30', 'blank': 'True'}),
      'password': ('django.db.models.fields.CharField', [], {'max_length': '128'}),
      'user_permissions': ('django.db.models.fields.related.ManyToManyField', [], {'to': "orm['auth.Permission']", 'symmetrical': 'False', 'blank': 'True'}),
      'username': ('django.db.models.fields.CharField', [], {'unique': 'True', 'max_length': '30'})
    },
    'contenttypes.contenttype': {
      'Meta': {'ordering': "('name',)", 'unique_together': "(('app_label', 'model'),)", 'object_name': 'ContentType', 'db_table': "'django_content_type'"},
      'app_label': ('django.db.models.fields.CharField', [], {'max_length': '100'}),
      'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
      'model': ('django.db.models.fields.CharField', [], {'max_length': '100'}),
      'name': ('django.db.models.fields.CharField', [], {'max_length': '100'})
    },
    'django_anyvcs.grouprights': {
      'Meta': {'unique_together': "(('repo', 'group'),)", 'object_name': 'GroupRights', 'db_table': "'anyvcs_grouprights'"},
      'created': ('django.db.models.fields.DateTimeField', [], {'auto_now_add': 'True', 'null': 'True', 'blank': 'True'}),
      'group': ('django.db.models.fields.related.ForeignKey', [], {'to': "orm['auth.Group']"}),
      'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
      'last_modified': ('django.db.models.fields.DateTimeField', [], {'auto_now': 'True', 'null': 'True', 'blank': 'True'}),
      'repo': ('django.db.models.fields.related.ForeignKey', [], {'to': "orm['django_anyvcs.Repo']"}),
      'rights': ('django.db.models.fields.CharField', [], {'default': "'rw'", 'max_length': '2'})
    },
    'django_anyvcs.repo': {
      'Meta': {'object_name': 'Repo', 'db_table': "'anyvcs_repo'"},
      'created': ('django.db.models.fields.DateTimeField', [], {'auto_now_add': 'True', 'null': 'True', 'blank': 'True'}),
      'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
      'last_modified': ('django.db.models.fields.DateTimeField', [], {'auto_now': 'True', 'null': 'True', 'blank': 'True'}),
      'name': ('django.db.models.fields.CharField', [], {'unique': 'True', 'max_length': '100', 'db_index': 'True'}),
      'path': ('django.db.models.fields.CharField', [], {'unique': 'True', 'max_length': '100', 'blank': 'True'}),
      'public_read': ('django.db.models.fields.BooleanField', [], {'default': 'False'}),
      'public_rights': ('django.db.models.fields.CharField', [], {'default': "'-'", 'max_length': '2'}),
      'vcs': ('django.db.models.fields.CharField', [], {'default': "'git'", 'max_length': '3'})
    },
    'django_anyvcs.userrights': {
      'Meta': {'unique_together': "(('repo', 'user'),)", 'object_name': 'UserRights', 'db_table': "'anyvcs_userrights'"},
      'created': ('django.db.models.fields.DateTimeField', [], {'auto_now_add': 'True', 'null': 'True', 'blank': 'True'}),
      'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
      'last_modified': ('django.db.models.fields.DateTimeField', [], {'auto_now': 'True', 'null': 'True', 'blank': 'True'}),
      'repo': ('django.db.models.fields.related.ForeignKey', [], {'to': "orm['django_anyvcs.Repo']"}),
      'rights': ('django.db.models.fields.CharField', [], {'default': "'rw'", 'max_length': '2'}),
      'user': ('django.db.models.fields.related.ForeignKey', [], {'to': "orm['auth.User']"})
    }
  }

  complete_apps = ['django_anyvcs']
